"""Utilities for type checking and type hinting."""
from typing import Union

from shapely.geometry import (
    GeometryCollection,
    LinearRing,
    LineString,
    MultiLineString,
    MultiPoint,
    MultiPolygon,
    Point,
    Polygon,
)

# This is imported in the other modules.
GeometryCollection

SubDomain = Union[Polygon, MultiPolygon]
Level = int

Vector = tuple[float, float]


SHAPELY_GEOMETRIES = (
    Point,
    LineString,
    Polygon,
    MultiPoint,
    MultiLineString,
    MultiPolygon,
    GeometryCollection,
    LinearRing,
)

ShapelyGeometry = (
    Point
    | LineString
    | Polygon
    | MultiPoint
    | MultiLineString
    | MultiPolygon
    | GeometryCollection
    | LinearRing
)



"""
This is the main module of `pargeo`.

The module provides the `Domain` class, which can be used to generate complex
two-dimensional geometries, also called domains. 

The `Domain` class simplifies the process of sequentially modifing the domain
by adding geometries, in form of shapely polygons, to some initial domain.

The polygons that are added to the domain are equipped with a level, which is used 
to introduce a hierarchy between them. The higher the level of the polygon is,
the more visible it is, i.e. it covers the polygons with a lower level.

Additionally, holes can be defined, holes are levels, and polygons whose level
is in holes are not visible, basically they are holes in the domain.

Example:

>>> from pargeo.domain import Domain
>>> from shapely.geometry import Polygon
<BLANKLINE>
>>> domain = Polygon([(0, 0), (1, 0), (1, 1), (0, 1)])
>>> domain = Domain(domain, holes={1})
>>> subdomain = Polygon(...) # Some shapely polygon
>>> domain.add(subdomain, level=1)

"""

from __future__ import annotations

from collections import UserDict
from typing import Any, Mapping, Protocol

import matplotlib.pyplot as plt

from shapely import GeometryCollection

from pargeo.utils.constants import BACKGROUND_LEVEL, DEFAULT_GRID_SIZE
from pargeo.utils.plot_utils import plot_legend, plot_polygon, DefaultColors, Color
from pargeo.utils.typing_utils import Level, MultiPolygon, Polygon, SubDomain


class Transform(Protocol):
    """
    The Transform protocol represents a callable that takes a SubDomain, a Level,
    a Domain, and any number of additional keyword arguments, and returns a SubDomain.

    This is used to define the interface for transformations that can be applied to a SubDomain.
    """

    def __call__(
        self, subdomain: SubDomain, level: Level, domain: Domain, **kwargs
    ) -> SubDomain:
        """
        Applies the transformation to the given SubDomain.

        Args:
            subdomain (SubDomain): The SubDomain to transform.
            level (Level): The Level of the subdomain.
            domain (Domain): The underlying domain.
            **kwargs: Additional keyword arguments for the transformation.

        Returns:
            SubDomain: The transformed SubDomain.
        """
        ...


class Constraint(Protocol):
    """
    The Constraint protocol represents a callable that takes a SubDomain, a Level,
    a Domain, and any number of additional keyword arguments, and returns a boolean.

    This is used to define the interface for constraints that can be applied to a SubDomain.
    """

    def __call__(
        self, subdomain: SubDomain, level: Level, domain: Domain, **kwargs
    ) -> bool:
        """
        Applies the constraint to the given SubDomain.

        Args:
            subdomain (SubDomain): The SubDomain to apply the constraint to.
            level (Level): The Level of the SubDomain.
            domain (Domain): The underlying domain.
            **kwargs: Additional keyword arguments for the constraint.

        Returns:
            bool: True if the SubDomain satisfies the constraint, False otherwise.
        """
        ...


class Domain:
    """Class for the creation of complex two-dimensional domains.

    A domain is a two-dimensional geometry, which can be modified by adding
    subdomains to it. The subdomains are shapely polygons, or multipolygons,
    and can be added using the `add` method. The subdomains are equipped with
    a level, which is used to introduce a hierarchy between them.

    Subdomains with a higher level will be cut out from those with a lower level.
    Visually speaking the higher level subdomains are more visible and cover the
    lower level subdomains. The lowest level is the `Domain.background_level`.

    Subdomains with the same level are merged together.
    """

    # Level of the background.
    background_level = BACKGROUND_LEVEL

    def __init__(
        self,
        background: SubDomain,
        holes: set[Level] = set(),
        grid_size: float = DEFAULT_GRID_SIZE,
    ) -> None:
        """Initialize the domain.

        Args:
            background (SubDomain): This subdomain is the background of the domain
                and is associated with the level `Domain.background_level`.

            holes (set[Level], optional): Set of levels that are considered as holes.
                Can be changed at any time using the `set_holes` method. Defaults to `set()`.

            grid_size (float, optional): Used `grid_size` for shapely geometry manipulating operations.
        """
        # Make sure background is a SubDomain
        if not isinstance(background, (Polygon, MultiPolygon)):
            raise ValueError(
                f"`background` must be a shapely Polygon or MultiPolygon, but is {type(background)}."
            )
        # Internally only work with MultiPolygons
        if isinstance(background, Polygon):
            background = MultiPolygon([background])
        self.__profile = background
        self.__holes = holes  # Can be changed at any time.
        self.__grid_size = grid_size
        self.__level_to_subdomain = {Domain.background_level: background}

    @property
    def level_to_subdomain(self) -> dict[Level, SubDomain]:
        """Dictionary mapping each level to its corresponding subdomain."""
        return self.__level_to_subdomain

    @property
    def holes(self) -> set[Level]:
        """Set of levels that are considered as holes."""
        return self.__holes

    def set_holes(self, new_holes: set[Level]):
        """Set `self.holes`."""
        self.__holes = new_holes

    @property
    def grid_size(self) -> float:
        """Used `grid_size` for shapely geometry manipulating operations."""
        return self.__grid_size

    @property
    def profile(self) -> SubDomain:
        """The profile of the domain.

        If the added subdomains are clipped to the background, then the profile
        matches the background. Otherwise it can be bigger.
        """
        return self.__profile

    @property
    def levels(self) -> list[Level]:
        """Present levels in sorted order."""
        return sorted(self.__level_to_subdomain.keys())

    @property
    def subdomains(self) -> list[SubDomain]:
        """Return a list of all subdomains."""
        subdomains = [subdomain for subdomain, _ in self.as_list()]
        return subdomains

    def add_subdomain(
        self,
        subdomain: SubDomain,
        level: Level,
        transform: Transform | None = None,
        constraint: Constraint | None = None,
        clip: bool = True,
        **kwargs: Any,
    ) -> bool:
        """
        Adds a subdomain to the domain.

        This method performs the following steps before potentially adding the subdomain:
            1. Transforms the incoming subdomain using `transform`.
            2. If clip is `True`, clips the subdomain to the domain's profile.
            3. If `constraint` are given, only proceede if these are met.
            4. Add the subdomain.

        Args:
            subdomain (SubDomain): The subdomain to be added.
            level (Level): The associated level of the subdomain.
            transform (Transform): A function to transform the subdomain before adding it.
                It will call `transform(subdomain, level, self)`.
            constraint (Constraint): A function to check if the subdomain should be added.
                It will call `constraints(subdomain, level, self)`.
            clip (bool): If `True`, clips the subdomain to the background.

        Note: If `level == Domain.background_level` and `clip == True`, then the
            background might be modified.

        Returns:
            bool: True if the subdomain was added, False otherwise.
        """
        if level < Domain.background_level:
            raise ValueError(
                f"`level` must be greater-equal to {Domain.background_level}."
            )

        # Apply transform if given
        if transform is not None:
            transform_kwargs = self._extract_kwargs(kwargs, "transform_")
            transform_kwargs["clip"] = clip
            subdomain = transform(subdomain, level, self, **transform_kwargs)

        # Clip the transformed subdomain to the background
        if clip:
            subdomain = subdomain.intersection(self.profile, grid_size=self.grid_size)
        else:
            profile = self.profile.union(subdomain, grid_size=self.grid_size)
            if isinstance(profile, Polygon):
                profile = MultiPolygon([profile])
            self.__profile = profile

        # Make sure subdomain is a MultiPolygon with positive area polygons
        if isinstance(subdomain, (GeometryCollection, MultiPolygon)):
            subdomain = MultiPolygon(
                [p for p in subdomain.geoms if p.area > self.grid_size]
            )

        # Make sure subdomain is not empty
        if subdomain.area < self.grid_size:
            return False

        # TODO: Apply segmentize strategy

        # Check constraints if given
        if constraint is not None:
            constraint_kwargs = self._extract_kwargs(kwargs, "constraint_")
            if not constraint(subdomain, level, self, **constraint_kwargs):
                return False

        # Update `self.__level_to_subdomain`
        level_to_subdomain = dict()
        for running_level, running_subdomain in self.__level_to_subdomain.items():
            if not subdomain.intersects(running_subdomain):
                # running_subdomain does not intersect with `subdomain` so it can be
                # added directly to `level_to_subdomain`.
                level_to_subdomain[running_level] = running_subdomain
                continue

            # `running_subdomain` intersects with `subdomain`
            # Both might need to be updated
            if level != running_level:
                higher_subdomain = (
                    subdomain if level > running_level else running_subdomain
                )
                lower_subdomain = (
                    subdomain if level < running_level else running_subdomain
                )

                # Cut out the higher level subdomain from the lower level subdomain
                lower_subdomain = lower_subdomain.difference(
                    higher_subdomain, grid_size=self.grid_size
                )

                # After shapely.difference, lower_subdomain can be a GeometryCollection.
                if isinstance(lower_subdomain, GeometryCollection):
                    lower_subdomain = MultiPolygon(
                        [p for p in lower_subdomain.geoms if p.area > self.grid_size]
                    )

                # If the lower subdomain is approximately empty, it is not added.
                # If subdomain = lower_subdomain, then the subdomain is not added.
                if lower_subdomain.area < self.grid_size:
                    if level < running_level:
                        return False
                    continue

                # Add the intersection points, this is needed to create delauny
                # triangulations by GMSH.
                intersection_ = higher_subdomain.intersection(
                    lower_subdomain, grid_size=self.grid_size
                )

                # `intersection` can be a GeometryCollection, if so, iterate over the
                # geoms and add them one by one to the higher_subdomain.
                if isinstance(intersection_, GeometryCollection):
                    for geom in intersection_.geoms:
                        # TODO: Check why geom can be a `shapely.geometry.Polygon`.
                        # This happens sometimes, and the `Polygon` has some approximate area 0.

                        # Only add shapely geometries with area == 0, e.g. Points
                        if geom.area == 0:
                            # Note `geom.area == 0` is needed to remove `Polygons` that,
                            # w.r.t. the given `grid_size` are not visible, *but* if a new
                            # `Polygon` is added with some epsilon area, then this needs
                            # to be removed, which happens in the next if statement.
                            higher_subdomain = higher_subdomain.union(
                                geom, grid_size=self.grid_size
                            )
                            if isinstance(higher_subdomain, GeometryCollection):
                                higher_subdomain = MultiPolygon(
                                    [
                                        poly
                                        for poly in higher_subdomain.geoms
                                        if poly.area > self.grid_size
                                    ]
                                )
                else:
                    higher_subdomain = higher_subdomain.union(
                        intersection_, grid_size=self.grid_size
                    )
                # TODO: Is it possible that the union in the upper if statement
                # is again a `Polygon`? If so, then we need to convert it.
                if isinstance(higher_subdomain, Polygon):
                    higher_subdomain = MultiPolygon([higher_subdomain])

                # TODO: Is this necessary?
                if isinstance(higher_subdomain, GeometryCollection):
                    higher_subdomain = MultiPolygon(
                        [p for p in higher_subdomain.geoms if p.area > self.grid_size]
                    )

                # Update  `subdomain` and `level_to_subdomain`
                if running_level < level:
                    assert not isinstance(lower_subdomain, GeometryCollection)
                    level_to_subdomain[running_level] = lower_subdomain
                    subdomain = higher_subdomain  # subdomain grows

                else:  # running_level > level
                    assert not isinstance(higher_subdomain, GeometryCollection)
                    level_to_subdomain[running_level] = higher_subdomain
                    subdomain = lower_subdomain  # subdomain shrinks

        # Merge the subdomain with the multipolygon of the same level.
        if level in self.__level_to_subdomain:
            subdomain = subdomain.union(
                self.__level_to_subdomain[level], grid_size=self.grid_size
            )
            # The shapely.union might create a GeometryCollection.
            # Appearently, polygon.union(multipolygon) can create
            # a GeometryCollection with a multiLineString.
            if isinstance(subdomain, GeometryCollection):
                subdomain = MultiPolygon(
                    [p for p in subdomain.geoms if p.area > self.grid_size]
                )

        if isinstance(subdomain, Polygon):
            subdomain = MultiPolygon([subdomain])

        if not isinstance(subdomain, MultiPolygon):
            raise ValueError(
                f"`subdomain` is not a MultiPolygon, but a {type(subdomain)}."
            )

        level_to_subdomain[level] = subdomain
        self.__level_to_subdomain = level_to_subdomain
        return True

    def plot(
        self,
        title: str = "Domain",
        color_holes: bool = False,
        color_hole_boundaries: bool = True,
        make_legend: bool = True,
        safe_file: str | None = None,
        color_map: Mapping[Level, Color] | None = None,
    ) -> None:
        """Plot the domain.

        Args:
            title (str, optional): Title of the plot. Defaults to "Domain".
            color_holes (bool, optional): If `True`, the holes are colored. Defaults to `False`.
            color_hole_boundaries (bool, optional): If `True`, the boundaries of the holes are colored.
                Defaults to `True`.
            make_legend (bool, optional): If `True`, a legend is added to the plot. Defaults to `True`.
            safe_file (str, optional): If given, the plot is saved to the file. Defaults to `None`.
            color_map (Mapping[Level, Color], optional): A mapping of levels to colors.
                If `None`, the default color map is used. Defaults to `None`.
        """
        if color_map is None:
            colormap = DefaultColors.get_color_map(
                self.levels, self.holes, color_holes, self.background_level
            )

        def plot_tree_rec(node: Node) -> None:
            """Recursively plot the domain."""
            if not node.is_root:
                node.plot(colormap, color_hole_boundaries, self.holes)

            for child in node["children"]:
                plot_tree_rec(child)

        plt.title(title)
        if make_legend:
            plot_legend(self.holes, colormap, Domain.background_level)
        plot_tree_rec(self.as_tree().root)
        if safe_file is not None:
            plt.savefig(safe_file, dpi=300)
        plt.show()

    def as_list(self) -> list[tuple[Polygon, Level]]:
        """Return a list of all subdomains with their levels."""
        polyongs_levels = []
        for level, running_polygon in self.__level_to_subdomain.items():
            if isinstance(running_polygon, MultiPolygon):
                for polygon in running_polygon.geoms:
                    polyongs_levels.append((polygon, level))
            else:
                polyongs_levels.append((running_polygon, level))
        return polyongs_levels

    def as_tree(self) -> InclusionTree:
        """Represent the domain as a tree.

        Each node of the tree represents a subdomain of the domain. If node A
        is a parent of node B, then the subdomain of node A includes the subdomain
        of node B.
        """
        root_node = Node(
            level=Domain.background_level - 1, polygon=Polygon(), children=[]
        )
        tree = InclusionTree(root=root_node)

        for polygon, level in self.as_list():
            tree.add(polygon, level)
        return tree

    def _extract_kwargs(self, kwargs: dict[str, Any], prefix: str) -> dict[str, Any]:
        """Extract the kwargs that start with `prefix`."""
        _kwargs_id = prefix
        _n = len(_kwargs_id)
        return {k[_n]: v for k, v in kwargs.items() if k.startswith(_kwargs_id)}


class InclusionTree:
    """Represents a `Domain` object as a tree.

    The tree is used to plot the domain.

    The tree is a directed acyclic graph, where the root is just an auxiliary
    node. Each (non-root) node represents a polygon, with a level. All polygons
    represented by the nodes are disjoint. The nodes represent the polygons in
    the flattened domain.

    This tree encodes the semi-ordering of the polygons, that is given by the
    inclusion relation. This means, a directed edge from node A to node B exists
    if and only if the polygon represented by node A includes the polygon of node B.
    Note: The polygon of the exterior of the polygon (spoken in terms of shapely) is
    taken into account for the inclusion.
    """

    def __init__(self, root: Node) -> None:
        root.is_root = True
        self.__root = root

    @property
    def root(self) -> Node:
        return self.__root

    def add(self, polygon: Polygon, level: Level) -> None:
        """
        Recursively add a node to the tree.

        This function adds a node to the tree by finding its correct position in relation to the parent node.
        The node is added as a child of the parent node if the parent node includes the node.
        If the node includes a child of the parent node, the child is removed from the parent node's children
        and added to the node's children. The depth of the node and any affected children is updated accordingly.

        Args:
            parent: The parent node.
            node: The node to be added.

        Returns:
            None
        """

        def add_node_rec(parent: Node, node: Node) -> None:
            """Recursively add a node to the tree."""
            for child in parent["children"]:
                if child.includes(node):
                    node.depth = child.depth + 1
                    add_node_rec(child, node)
                    return

                elif node.includes(child):
                    node.depth = child.depth
                    child.depth = node.depth + 1
                    node["children"].append(child)
                    parent["children"].remove(child)
                    add_node_rec(parent, node)
                    return
                else:
                    continue
            node.depth = parent.depth + 1
            parent["children"].append(node)

        node = Node(level=level, polygon=polygon)
        add_node_rec(self.__root, node)

    def show(self) -> None:
        """Print the tree recursively."""

        def show_rec(node: Node, depth: int) -> None:
            """Recursively print each node."""
            if not node.is_root:
                print("    " * (depth - 1), node)
            for child in node["children"]:
                show_rec(child, depth + 1)

        show_rec(self.__root, depth=0)


class Node(UserDict):
    """Node of the inclusion tree, representing a polygon.

    Nodes are semi-ordered by the inclusion relation.
    """

    def __init__(self, level: Level, polygon: Polygon, children=None):
        super().__init__()
        self.data = {
            "level": level,
            "polygon": polygon,
            "children": children if children is not None else [],
        }
        self.is_root = False
        self.depth = 0

    def includes(self, node: Node) -> bool:
        """Check if this node includes the incoming node."""
        return Polygon(self["polygon"].exterior).contains(node["polygon"])

    def __repr__(self) -> str:
        return f"Node(level={self['level']}, depth={self.depth}), n_children={len(self['children'])}"

    def plot(
        self,
        colormap: Mapping[Level, Color],
        color_hole_boundaries: bool,
        holes: set[int],
        show: bool = False,
        boundary_color: Color | None = None,
    ):
        """Plot the node."""
        is_hole = self.data["level"] in holes
        color = colormap[self.data["level"]]
        polygon = self.data["polygon"]
        if boundary_color is None:
            boundary_color = DefaultColors.boundary

        if not is_hole:
            plot_polygon(polygon, color, boundary_color, "-", 1.0, show)
        else:
            if color_hole_boundaries:
                plot_polygon(polygon, color, boundary_color, "--", 0.5, show)
            else:
                plot_polygon(polygon, color, "none", "-", 1.0, show)



import math
from typing import Callable

from shapely import box as shapely_box
from shapely import union_all
from shapely.affinity import translate as shapely_translate

from pargeo.utils.constants import DEFAULT_GRID_SIZE
from pargeo.utils.typing_utils import (
    GeometryCollection,
    MultiPolygon,
    Polygon,
    SubDomain,
    Vector,
)


def discretize_ellipse(
    midpoint: Vector, axis: Vector, angle: float, refs: int
) -> Polygon:
    """Discretize the ellipse.

    Args:
        midpoint: Ellipse's midpoint.
        axis: Ellipse's axis.
        angle: Rotation angle of the ellipse in radiant.
        refs: Number of discretization points.

    Return:
        List of boundary points in ccw order.
    """
    angles = [i * 2 * math.pi / refs for i in range(refs)]

    transformed_coords = rotate_counterclockwise(
        points=[polar_to_cartesian(angle, axis[0], axis[1]) for angle in angles],
        angle=angle,
    )
    mx, my = midpoint

    return Polygon([(x + mx, y + my) for x, y in transformed_coords])


def trigonometric_function(
    coefficients: list[Vector], scale: float
) -> Callable[[float], float]:
    """Returns the exponential of a trigonometric function.

    Args:
        coefficients: Coefficients [(a_i, b_i)_i] of the trigonometric function.
        scale: Scaling factor.

    The trigonometric function f is defined as:

        `f(x) = scale * exp(f_0(x) + ... + f_{n-1}(x))`

    where the basis functions f_i are defined as:

        `f_i(x) := a_i*cos(i*x) + b_i*sin(i*x)`
    """
    return lambda x: scale * math.exp(
        sum(
            a * math.cos(i * x) + b * math.sin(i * x)
            for i, (a, b) in enumerate(coefficients)
        )
    )


def rotate_counterclockwise(points: list[Vector], angle: float) -> list[Vector]:
    """Rotates points counter-clockwise around (0,0)."""
    cos_alpha = math.cos(angle)
    sin_alpha = math.sin(angle)
    return [
        (cos_alpha * x - sin_alpha * y, sin_alpha * x + cos_alpha * y)
        for x, y in points
    ]


def polar_to_cartesian(
    angle: float,
    rad_x: float,
    rad_y: float | None = None,
) -> Vector:
    """Coordinate transform, from polar to cartesian.

    Args:
        angles: List of angles.
        radii_x: List of radii, for the x-coordinate.
        radii_y: List of radii, for the y-coordinate.
    """
    if rad_y is None:
        rad_y = rad_x
    return (math.cos(angle) * rad_x, math.sin(angle) * rad_y)


def repeat(
    subdomain: SubDomain,
    repeat_info: tuple[Vector, float, int, bool],
    repeat_info_2: tuple[Vector, float, int, bool] | None = None,
    clipping_plane: SubDomain | None = None,
    grid_size: float = DEFAULT_GRID_SIZE,
) -> SubDomain:
    """Repeat the subdomain in one or two directions."""
    v_dir, v_scalar, v_rep, v_do_opposite = repeat_info
    if repeat_info_2 is not None:
        w_dir, w_scalar, w_rep, w_do_opposite = repeat_info_2
    else:
        w_dir, w_scalar, w_rep, w_do_opposite = (0, 0), 0, 0, False
    shifted_polygons = []

    bounds = subdomain.bounds
    i_range = range(v_rep) if not v_do_opposite else range(-v_rep + 1, v_rep)
    j_range = range(w_rep) if not w_do_opposite else range(-w_rep + 1, w_rep)
    for i in i_range:
        for j in j_range:
            x_off = i * v_scalar * v_dir[0]
            y_off = i * v_scalar * v_dir[1]
            x_off += j * w_scalar * w_dir[0]
            y_off += j * w_scalar * w_dir[1]

            # If clip to clipping_plane if given.
            if clipping_plane is not None:
                bounds_shifted = (
                    bounds[0] + x_off,
                    bounds[1] + y_off,
                    bounds[2] + x_off,
                    bounds[3] + y_off,
                )
                # Check if the shifted bounds intersect with the clipping plane.
                if not shapely_box(*bounds_shifted).intersects(clipping_plane):
                    continue

            shifted_polygons.append(
                shapely_translate(subdomain, xoff=x_off, yoff=y_off)
            )

    # Merge all shifted polygons.
    repeated_polygon = union_all(shifted_polygons)

    # Clip to domain.
    if clipping_plane is not None:
        repeated_polygon = repeated_polygon.intersection(
            clipping_plane, grid_size=grid_size
        )
        if isinstance(repeated_polygon, GeometryCollection):
            repeated_polygon = MultiPolygon(
                [poly for poly in repeated_polygon if poly.area > grid_size]
            )

    return repeated_polygon



import pargeo.constraint as constraint
import pargeo.geometry as geometry
import pargeo.transform as transform
from pargeo.domain import Domain
from pargeo.utils.gmsh_utils import write_geo, write_mesh
from pargeo.utils.plot_utils import plot_subdomain

__all__ = [
    "Domain",
    "geometry",
    "transform",
    "constraint",
    "write_geo",
    "write_mesh",
    "plot_subdomain",
]



"""
This module defines constraints for geometric objects in the pargeo library.

It includes an abstract base class `Constraint` that defines the interface for all constraints,
and `DistanceConstraint`.

`DistanceConstraint` allows setting distance constraints between different geometric objects.
These objects can be domain multipolygons, addressed by their level or shapely Geometry objects. 
The distances can be set between any two objects, or to the boundary of an object. 
The constraints are stored in three dictionaries, one for distances between domain multipolygons, one for distances to
the boundaries of domain multipolygons, and one for distances to shapely Geometry objects.


"""
from itertools import product
from typing import Any, Dict, Literal, overload

from pargeo.domain import Constraint, Domain
from pargeo.utils.typing_utils import (
    SHAPELY_GEOMETRIES,
    Level,
    ShapelyGeometry,
    SubDomain,
)

IntraConstraintDict = Dict[tuple[Level | Literal["all"], Level | Literal["all"]], float]

InterConstraintDict = Dict[tuple[Level | Literal["all"], ShapelyGeometry], float]


class DistanceConstraint(Constraint):
    """Distance constraint."""

    def __init__(self) -> None:
        """Initialize the distance constraint.

        The distance constraint are stored in three dictionaries:
        - `__polygon_constraints`: Distance constraints between the domain's subdomains.
        - `__boundary_constraints`: Distance constraints between domains' subdomain' boundaries.
        - `__geoms_constraints`: Distance constraints to arbitrary shapely Geometry objects.
        """
        self.__polygon_constraints: IntraConstraintDict = dict()

        self.__boundary_constraints: IntraConstraintDict = dict()

        self.__geoms_constraints: InterConstraintDict = dict()

    def set_distance(
        self,
        obj_1: Level | Literal["all"],
        obj_2: Level | Literal["all"] | ShapelyGeometry,
        distance: float,
        to_boundary: bool = False,
    ):
        """Set distance constraint.

        If one of the objects is `"all"`, the distance constraint is set for all levels.

        Args:
            distance (float): Minimal distance between `obj_1` and `obj_2`.
            obj_1 (Level | Literal["all"]): The first object.
            obj_2 (Level | Literal["all"] | ShapelyGeometry): The second object.
            to_boundary (bool): If `True`, the distance is measured to the boundary. Defaults to `False`.
        """
        # Validate input.
        if not (isinstance(obj_1, int) or obj_1 == "all"):
            raise ValueError("`obj_1` must be a level or 'all'")

        if not (isinstance(obj_2, (int, SHAPELY_GEOMETRIES)) or obj_2 == "all"):
            raise ValueError("`obj_1` must be a level or 'all' or a shapely Geometry.")

        # Set distance constraints to shapely Geometry objects.
        if isinstance(obj_2, SHAPELY_GEOMETRIES):
            self.__geoms_constraints[(obj_1, obj_2)] = distance
            return

        # Set intra-domain distance constraints.
        # Distance constraint to the whole polygon.
        if not to_boundary:
            self.__polygon_constraints[(obj_1, obj_2)] = distance
            if obj_1 != obj_2:
                self.__polygon_constraints[(obj_2, obj_1)] = distance

        else:  # Distance constraint to the boundary of the polygon.
            self.__boundary_constraints[(obj_1, obj_2)] = distance
            if obj_1 != obj_2:
                self.__boundary_constraints[(obj_2, obj_1)] = distance

    def __call__(
        self, subdomain: SubDomain, level: Level, domain: Domain, **kwargs: Any
    ) -> bool:
        """Compute the distance constraint.

        Args:
            subdomain (SubDomain): The subdomain to check.
            level (Level): The level of the subdomain.
            domain (Domain): The domain to check against.
        """
        # Get all present levels of `domain`.
        levels = domain.levels
        grid_size = domain.grid_size
        background_level = domain.background_level
        if level not in levels:
            levels.append(level)

        # MultiPolygon collision check, ignores background.
        dist_dict_poly = self.__all_to_lvl(
            self.__polygon_constraints, levels, True, background_level
        )
        lvls_dists = [
            (lvl2, dist)
            for (lvl1, lvl2), dist in dist_dict_poly.items()
            if lvl1 == level
        ]
        for lvl, dist in lvls_dists:
            # Check collision level with lvl.
            if lvl in domain.level_to_subdomain:
                subdomain_2 = domain.level_to_subdomain[lvl]
                if subdomain.distance(subdomain_2) < dist - grid_size:
                    return False

        # MultiPolygon boundary collision check, DOESN'T ignore the backgroud.
        dist_dict_boundary = self.__all_to_lvl(
            self.__boundary_constraints, levels, False, background_level
        )
        lvls_dists = [
            (lvl2, dist)
            for (lvl1, lvl2), dist in dist_dict_boundary.items()
            if lvl1 == level
        ]
        for lvl, dist in lvls_dists:
            if lvl in domain.level_to_subdomain:
                subdomain_2 = domain.level_to_subdomain[lvl]
                if subdomain.distance(subdomain_2.boundary) < dist - grid_size:
                    return False

        # Geometry collision check, ignores the background.
        dist_geoms = self.__all_to_lvl(
            self.__geoms_constraints, levels, True, background_level
        )
        geoms_dists = [
            (geom, dist) for (lvl1, geom), dist in dist_geoms.items() if lvl1 == level
        ]
        for geom, dist in geoms_dists:
            if subdomain.distance(geom) < dist - grid_size:
                return False

        # All collision checks passed.
        return True

    def show(self):
        """Print the distance constraints."""
        print("Distance constraints:")
        print("Subdomains:")
        for (lvl1, lvl2), dist in self.__polygon_constraints.items():
            print(f"    {lvl1} - {lvl2}: {dist}")
        print("  Boundaries:")
        for (lvl1, lvl2), dist in self.__boundary_constraints.items():
            print(f"    {lvl1} - {lvl2}: {dist}")
        print("  Geometries:")
        for (lvl1, geom), dist in self.__geoms_constraints.items():
            print(f"    {lvl1} - {geom}: {dist}")

    @overload
    def __all_to_lvl(
        self,
        distance_dict: IntraConstraintDict,
        final_levels: list[Level],
        ignore_background: bool,
        background_level: Level,
    ) -> Dict[tuple[Level, Level], float]:
        ...

    @overload
    def __all_to_lvl(
        self,
        distance_dict: InterConstraintDict,
        final_levels: list[Level],
        ignore_background: bool,
        background_level: Level,
    ) -> Dict[tuple[Level, ShapelyGeometry], float]:
        ...

    def __all_to_lvl(
        self,
        distance_dict: IntraConstraintDict | InterConstraintDict,
        final_levels: list[Level],
        ignore_background: bool,
        background_level: Level,
    ) -> (
        Dict[tuple[Level, Level], float] | Dict[tuple[Level, ShapelyGeometry], float]
    ):  # TODO: Add type hints for return
        """Replace `all` with the actual levels."""
        reduced_distance_dict: Dict[tuple[Level, Level], float] | Dict[
            tuple[Level, ShapelyGeometry], float
        ] = dict()
        if ignore_background:

            def do_update(lvl1, lvl2, dist):
                return background_level not in (lvl1, lvl2) and (
                    lvl1,
                    lvl2 not in reduced_distance_dict
                    or reduced_distance_dict[lvl1, lvl2] < dist,
                )

        else:

            def do_update(lvl1, lvl2, dist):
                return (
                    lvl1,
                    lvl2 not in reduced_distance_dict
                    or reduced_distance_dict[lvl1, lvl2] < dist,
                )

        for (lvl_1, lvl_2), dist in distance_dict.items():
            match lvl_1, lvl_2:
                case "all", "all":
                    for lvl_1_, lvl_2_ in product(final_levels, repeat=2):
                        if do_update(lvl_1_, lvl_2_, dist):
                            reduced_distance_dict[lvl_1_, lvl_2_] = dist
                case "all", _:
                    for lvl in final_levels:
                        if do_update(lvl, lvl_2, dist):
                            reduced_distance_dict[(lvl, lvl_2)] = dist  # type: ignore
                case _, "all":
                    for lvl in final_levels:
                        if do_update(lvl_1, lvl, dist):
                            reduced_distance_dict[(lvl_1, lvl)] = dist  # type: ignore
                case _:
                    if do_update(lvl_1, lvl_2, dist):
                        reduced_distance_dict[(lvl_1, lvl_2)] = dist  # type: ignore

        return reduced_distance_dict



"""API for Gmsh.

This module provides an API for Gmsh. It allows the user to write a Gmsh file 
and create a mesh using Gmsh.
"""

from itertools import groupby
from pathlib import Path
from typing import NamedTuple, Sequence, TypedDict

import gmsh
from shapely import Polygon

from pargeo.domain import Domain

PHISICAL_DIMENSION = 2


class Point(NamedTuple):
    """Two dimensional point."""

    x: float
    y: float
    z: float
    lc: float
    tag: int


class Line(NamedTuple):
    """Line between two points."""

    start_tag: int
    end_tag: int
    tag: int


class CurveLoop(NamedTuple):
    """Closed loop of lines."""

    line_tags: Sequence[int]
    tag: int


class PlaneSurface(NamedTuple):
    """Surface, delimited by curve loops."""

    curve_loop_tags: Sequence[int]
    tag: int


class PhysicalGroup(NamedTuple):
    """Group of entities belonging to the same physical group."""

    entity_tags: Sequence[int]
    tag: int


class SurfaceLoop(NamedTuple):
    """Closed loop of surfaces."""

    surface_tags: Sequence[int]
    tag: int


class Volume(NamedTuple):
    """Volume, delimited by surface loops."""

    surface_loop_tags: Sequence[int]
    tag: int


class GmshEntities(TypedDict):
    """Entities needed to create a mesh using gmsh."""

    points: Sequence[Point]
    lines: Sequence[Line]
    curve_loops: Sequence[CurveLoop]
    plane_surfaces: Sequence[PlaneSurface]
    physical_groups: Sequence[PhysicalGroup]
    surface_loops: Sequence[SurfaceLoop]
    volumes: Sequence[Volume]


def write_msh_from_entities(
    gmsh_entities: GmshEntities,
    file_name: Path,
    write_geo: bool = True,
    correct_curve_loops: bool = False,
    save_all: bool = False,
) -> None:
    """
    Write a mesh .MSH file using GMSH.

    Args:
        gmsh_entities: Entities that represent the geometry.
        file_name: Path of the saved mesh, the extension is added automatically.
        write_geo: Whether or not the .GEO file is saved too.
        correct_line_loops: Apparently gmsh offeres the possibility to correct
            line loops. TODO: Check if it is useful.
        save_all: If False only mesh elements are saved that belong to some
            physical group, as long there is at least one physical group.
    """

    gmsh.initialize()
    gmsh.model.add(file_name.name)

    for point in gmsh_entities["points"]:
        gmsh.model.geo.addPoint(
            x=point.x, y=point.y, z=point.z, meshSize=point.lc, tag=point.tag
        )

    for line in gmsh_entities["lines"]:
        gmsh.model.geo.addLine(
            startTag=line.start_tag, endTag=line.end_tag, tag=line.tag
        )

    for curve_loop in gmsh_entities["curve_loops"]:
        gmsh.model.geo.addCurveLoop(
            curveTags=curve_loop.line_tags,
            tag=curve_loop.tag,
            reorient=correct_curve_loops,
        )

    for plane_surface in gmsh_entities["plane_surfaces"]:
        gmsh.model.geo.addPlaneSurface(
            wireTags=plane_surface.curve_loop_tags, tag=plane_surface.tag
        )

    # Add physical groups
    for physical_group in gmsh_entities["physical_groups"]:
        gmsh.model.geo.addPhysicalGroup(
            dim=PHISICAL_DIMENSION,
            tag=physical_group.tag,
            tags=physical_group.entity_tags,
        )

    gmsh.model.geo.synchronize()
    gmsh.model.mesh.generate(PHISICAL_DIMENSION)

    # Choose Dolfin compatible version
    gmsh.option.setNumber("Mesh.MshFileVersion", 2)

    if save_all:
        # Write all mesh entities
        gmsh.option.setNumber("Mesh.SaveAll", 1)

    # Write the mesh file
    gmsh.write(str(file_name.with_suffix(".msh")))

    if write_geo:
        # Write the geo file
        gmsh.write(str(file_name.with_suffix(".geo_unrolled")))

    gmsh.finalize()


def write_geo_from_entities(
    file_name: Path,
    gmsh_entities: GmshEntities,
    correct_curve_loops: bool = False,
):
    """Write a file readable by GMSH.

    The geo file can be used in gmsh.

    Args:
        file_name: Name of the file. The extension ".geo_unrolled" is added
            automatically.
        gmsh_entities: Entities that represent the geometry.
        correct_curve_loops: Apparently gmsh offeres the possibility to correct
            line loops. TODO: Check if it is useful.
    """
    gmsh.initialize()
    gmsh.model.add(file_name.name)

    for point in gmsh_entities["points"]:
        gmsh.model.geo.addPoint(
            x=point.x, y=point.y, z=point.z, meshSize=point.lc, tag=point.tag
        )

    for line in gmsh_entities["lines"]:
        gmsh.model.geo.addLine(
            startTag=line.start_tag, endTag=line.end_tag, tag=line.tag
        )

    for curve_loop in gmsh_entities["curve_loops"]:
        gmsh.model.geo.addCurveLoop(
            curveTags=curve_loop.line_tags,
            tag=curve_loop.tag,
            reorient=correct_curve_loops,
        )

    for plane_surface in gmsh_entities["plane_surfaces"]:
        gmsh.model.geo.addPlaneSurface(
            wireTags=plane_surface.curve_loop_tags, tag=plane_surface.tag
        )

    for physical_group in gmsh_entities["physical_groups"]:
        gmsh.model.geo.addPhysicalGroup(
            dim=PHISICAL_DIMENSION,
            tag=physical_group.tag,
            tags=physical_group.entity_tags,
        )
    gmsh.model.geo.removeAllDuplicates()

    gmsh.model.geo.synchronize()

    # Write geo file
    gmsh.write(str(file_name.with_suffix(".geo_unrolled")))
    gmsh.finalize()


def polygon_to_entities(
    polygon: Polygon,
    level: int,
    point_tag: int,
    line_tag: int,
    curve_loop_tag: int,
    plane_surface_tag: int,
):
    """Convert a shapely polygon to gmsh entities.

    Args:
        polygon: Shapely polygon.
        level: Physical group of the surface.
        point_tag: Lower bound for free point tags.
        line_tag: Lower bound for free line tags.
        curve_loop_tag: Lower bound for free curve_loop tags.
        plane_surface_tag: Lower bound for free surface tags.
    """
    # assert not bubble.is_hole
    points_total: list[Point] = []
    lines_total: list[Line] = []
    curve_loops: list[CurveLoop] = []

    physical_group_tag = level + 1  # +1 because physical group tags start at 1
    # ==================================================
    # Write exterior loop
    # ==================================================
    assert isinstance(polygon, Polygon)
    # Add points
    points_local: list[Point] = []
    for x, y in polygon.exterior.coords[:-1]:
        point = Point(x=x, y=y, z=0.0, lc=0.1, tag=point_tag)
        points_local.append(point)
        point_tag += 1

    # Add lines
    lines_local = []
    for i in range(len(points_local) - 1):
        line = Line(
            start_tag=points_local[i].tag, end_tag=points_local[i + 1].tag, tag=line_tag
        )
        lines_local.append(line)
        line_tag += 1
    line = Line(
        start_tag=points_local[-1].tag, end_tag=points_local[0].tag, tag=line_tag
    )
    lines_local.append(line)
    line_tag += 1

    points_total.extend(points_local)
    lines_total = lines_total + lines_local
    # Add curve loop
    curve_loop = CurveLoop(
        line_tags=[line.tag for line in lines_local], tag=curve_loop_tag
    )
    curve_loops.append(curve_loop)
    curve_loop_tag += 1

    # ==================================================
    # Write interior loops
    # ==================================================
    for loop in polygon.interiors:
        points_local = []
        lines_local = []
        for x, y in loop.coords[:-1]:
            point = Point(x=x, y=y, z=0.0, lc=0.1, tag=point_tag)
            points_local.append(point)
            point_tag += 1

        # Add lines
        for i in range(len(points_local) - 1):
            line = Line(
                start_tag=points_local[i].tag,
                end_tag=points_local[i + 1].tag,
                tag=line_tag,
            )
            lines_local.append(line)
            line_tag += 1
        line = Line(
            start_tag=points_local[-1].tag, end_tag=points_local[0].tag, tag=line_tag
        )
        lines_local.append(line)
        line_tag += 1

        # Add curve loop
        curve_loop = CurveLoop(
            line_tags=[line.tag for line in lines_local], tag=curve_loop_tag
        )
        curve_loops.append(curve_loop)
        curve_loop_tag += 1

        points_total = points_total + points_local
        lines_total = lines_total + lines_local

    # ==================================================
    # Write surface
    # ==================================================
    plane_surfaces = [
        PlaneSurface(
            curve_loop_tags=[loop.tag for loop in curve_loops], tag=plane_surface_tag
        )
    ]
    physical_groups = [
        PhysicalGroup(
            entity_tags=[plane_surface_tag],
            tag=physical_group_tag,
        )
    ]
    plane_surface_tag += 1

    entities: GmshEntities = {
        "points": points_total,
        "lines": lines_total,
        "curve_loops": curve_loops,
        "plane_surfaces": plane_surfaces,
        "physical_groups": physical_groups,
        "surface_loops": [],
        "volumes": [],
    }
    return (
        entities,
        point_tag,
        line_tag,
        curve_loop_tag,
        plane_surface_tag,
    )


def domain_to_entities(domain: Domain) -> GmshEntities:
    """Convert a domain to gmsh entities."""
    all_entities_list = []

    # TODO: Rename this to something more meaningful
    pargeo = domain.as_list()

    point_tag = 1
    line_tag = 1
    curve_loop_tag = 1
    plane_surface_tag = 1

    for polygon, level in pargeo:
        if level not in domain.holes:
            (
                gmsh_entities,
                point_tag,
                line_tag,
                curve_loop_tag,
                plane_surface_tag,
            ) = polygon_to_entities(
                polygon, level, point_tag, line_tag, curve_loop_tag, plane_surface_tag
            )
            all_entities_list.append(gmsh_entities)

    # Group all physical groups by tag
    physical_groups_merged = []
    physical_groups = [e["physical_groups"][0] for e in all_entities_list]
    physical_groups.sort(key=lambda x: x.tag)

    for tag, physical_group in groupby(physical_groups, lambda x: x.tag):
        entity_tags = [s.entity_tags[0] for s in physical_group]
        physical_groups_merged.append(PhysicalGroup(entity_tags=entity_tags, tag=tag))

    gmsh_entities_all: GmshEntities = {
        "points": [i for ent in all_entities_list for i in ent["points"]],
        "lines": [i for ent in all_entities_list for i in ent["lines"]],
        "curve_loops": [i for ent in all_entities_list for i in ent["curve_loops"]],
        "plane_surfaces": [
            i for ent in all_entities_list for i in ent["plane_surfaces"]
        ],
        "physical_groups": physical_groups_merged,
        "surface_loops": [],
        "volumes": [],
    }
    return gmsh_entities_all


def write_geo(
    domain: Domain,
    file_name: Path | str,
    correct_curve_loops: bool = False,
) -> None:
    """Write a Gmsh file.

    The file will be saved as a `.geo_unrolled` file that can be opened in Gmsh.

    Args:
        domain (Domain): The domain to be written to a file.
        file_name (Path | str): The file name to save the domain to.
        correct_curve_loops (bool, optional): Whether Gmsh should correct the curve loops.
            Defaults to False.
    """
    gmsh_entities = domain_to_entities(domain)
    write_geo_from_entities(Path(file_name), gmsh_entities, correct_curve_loops)


def write_mesh(
    domain: Domain,
    file_name: Path | str,
    write_geo: bool = True,
    correct_curve_loops: bool = False,
    save_all: bool = False,
) -> None:
    """Create and write mesh with Gmsh.

    Use the Gmsh software to create a mesh and write it to a `.MSH` file.

    Args:
        domain (Domain): The domain to be meshed.
        file_name (Path | str): The file name to save the mesh to. Suffix `.msh` will be added.
        write_geo (bool, optional): Whether to write a `.geo_unrolled` file. Defaults to True.
        correct_curve_loops (bool, optional): Whether to correct the curve loops. Defaults to False.
        save_all (bool, optional): Whether to save all files. Defaults to False. TODO: What does this mean?
    """
    gmsh_entities = domain_to_entities(domain)
    write_msh_from_entities(
        gmsh_entities,
        Path(file_name),
        write_geo,
        correct_curve_loops,
        save_all,
    )



"""Geometry module.

Utility module for creating geometries. The module provides a set of classes
for creating geometries such as rectangles, circles, ellipses, and star-like
geometries. These geometries can be discretized using the `to_polygon` method,
and used in the `Domain` class.

Example:

    >>> from pargeo.domain import Domain
    >>> from pargeo.geometry import Rectangle, Circle
    <BLANKLINE>
    >>> domain = Rectangle(center=(0, 0), width=1, height=1).to_polygon()
    >>> domain = Domain(domain, holes={1})
    >>> hole = Circle(center=(.5, .5), radius=.1).to_polygon()
    >>> domain.add(polygon=hole, level=1)

"""
import math
import random
from abc import ABC, abstractmethod

from pargeo.utils.geometry_utils import (
    discretize_ellipse,
    polar_to_cartesian,
    trigonometric_function,
)
from pargeo.utils.typing_utils import Polygon, Vector


class Geometry(ABC):
    """Abstract class for geometries."""

    @abstractmethod
    def to_polygon(self, *args, **kwargs) -> Polygon:
        """Return a shapely polygon representation of self."""
        pass


class Box(Geometry):
    """Box geometry."""

    def __init__(self, left_bottom: Vector, right_top: Vector):
        """Creates a rectangle.

        Args:
            left_bottom (tuple): The (x, y) coordinates of the left bottom corner.
            right_top (tuple): The (x, y) coordinates of the right top corner.
        """
        self.left_bottom = left_bottom
        self.right_top = right_top
        self.center = tuple(right_top[i] - left_bottom[i] for i in range(2))
        self.width = right_top[0] - left_bottom[0]
        self.height = right_top[1] - left_bottom[1]

    @classmethod
    def from_center(cls, center: Vector, width: float, height: float):
        """
        Creates a box from the center.

        Args:
            center (tuple): The (x, y) coordinates of the center of the box.
            width (float): The width of the box.
            height (float): The height of the box.

        Returns:
            Box: The created box object.
        """
        left_bottom = (center[0] - width / 2, center[1] - height / 2)
        right_top = (center[0] + width / 2, center[1] + height / 2)
        return cls(left_bottom, right_top)

    def to_polygon(self) -> Polygon:
        """Return the corners of the rectangle in counter-clockwise order."""
        p0 = self.left_bottom
        p1 = (self.right_top[0], self.left_bottom[1])
        p2 = self.right_top
        p3 = (self.left_bottom[0], self.right_top[1])

        return Polygon([p0, p1, p2, p3])


class NStar(Geometry):
    """NStar geometry."""

    def __init__(
        self,
        center: Vector,
        radius_in: float,
        radius_out: float,
        n_peaks: int,
        alpha: float = 0.0,
    ):
        """Make an NStar.

        Args:
            center: The center.
            radius_in: The inner radius.
            radius_out: The outer radius.
            N: Number of points.
            alpha: Angle offset. TODO: Explain this.
        """
        if radius_in > radius_out:
            raise ValueError(
                "The parameter setting radius_in > radius_out not allowed."
            )
        self._center = center
        self._inner_radius = radius_in
        self._outer_radius = radius_out
        self._n_peaks = n_peaks
        self.__alpha = alpha

    def to_polygon(self) -> Polygon:
        aux_angles = [2 * math.pi * i / self._n_peaks for i in range(self._n_peaks)]
        offset = math.pi / self._n_peaks

        angles_in = [angle + self.__alpha + offset for angle in aux_angles]
        angles_out = [angle + self.__alpha for angle in aux_angles]

        coords_in = [
            polar_to_cartesian(angle, self._inner_radius) for angle in angles_in
        ]
        coords_out = [
            polar_to_cartesian(angle, self._outer_radius) for angle in angles_out
        ]

        # Translate to center
        mx, my = self._center
        coords_in = [(x + mx, y + my) for x, y in coords_in]
        coords_out = [(x + mx, y + my) for x, y in coords_out]

        # Create a shapely polygon
        xy = []
        for xy_in, xy_out in zip(coords_in, coords_out):
            xy.append(xy_out)
            xy.append(xy_in)

        return Polygon(xy)


class StarLike(Geometry):
    """Abstract class for star-like geometries."""

    def __init__(self, center: Vector) -> None:
        self.center = center

    @abstractmethod
    def radius_at(self, angle: float) -> float:
        """Returns the distance at a given angle to the center."""
        raise NotImplementedError

    def to_polygon(self, refs: int) -> Polygon:
        """Return a shapely polygon representation of self."""
        return self.discretize(refs)

    def discretize(self, refs: int) -> Polygon:
        # TODO: also give angles as alternative.
        angles = [i * 2 * math.pi / refs for i in range(refs)]
        radii = [self.radius_at(angle) for angle in angles]
        mx, my = self.center

        coords = [polar_to_cartesian(angle, rad) for angle, rad in zip(angles, radii)]

        return Polygon([(x + mx, y + my) for x, y in coords])


class Circle(StarLike):
    def __init__(self, center: Vector, radius) -> None:
        self.radius = radius
        super().__init__(center)

    def radius_at(self, angle: float) -> float:
        return self.radius


class Ellipse(StarLike):
    def __init__(self, center: Vector, axis: Vector, angle: float = 0) -> None:
        self.axis = axis
        self.angle = angle
        super().__init__(center)

    def radius_at(self, angle: float) -> float:
        # TODO: Check if this is correct
        return math.sqrt(
            (self.axis[0] * math.cos(angle)) ** 2
            + (self.axis[1] * math.sin(angle)) ** 2
        )

    def discretize(self, refs) -> Polygon:
        return discretize_ellipse(self.center, self.axis, self.angle, refs)


class Stellar(StarLike):
    """Stellar bubble."""

    def __init__(
        self,
        center: Vector,
        radius: float,
        coefficient: list[tuple[float, float]] | None = None,
    ):
        """Creates a star-like geometry.

        Args:
            center: The center.
            radius: Approximately the average radius.
            coefficient: numpy array of shape (n, 2) with n > 0.
                Hint: for smooth stellar holes, you can çhoose k from {.1, .2}
                and then `coeffients = np.random.uniform(-k, k, (int(1/k), 2))`
        """
        self.center = center
        self.radius = radius
        if coefficient is None:
            coefficient = self.generate_coefficients()

        self.coefficient = coefficient
        self.trig_fun = trigonometric_function(self.coefficient, self.radius)

    def radius_at(self, angle: float) -> float:
        return self.trig_fun(angle)

    def generate_coefficients(self, k: float = 0.2) -> list[tuple[float, float]]:
        """Generate coefficients for the stellar function.

        Args:
            k: Scaling factor.
        """
        j = int(1 / k)
        coeffs_ = [random.uniform(-k, k) for _ in range(j * 2)]

        return list(zip(coeffs_[:j], coeffs_[j:]))


class RainDrop(StarLike):
    """Raindrop geometry."""

    def __init__(self, center: Vector, a: float, scale: float):
        super().__init__(center)
        if not (0.5 <= a <= 1):
            raise ValueError(f"Value of a must be between 0.5 and 1, but got {a}.")
        self._a = a
        self._scale = scale

    def radius_at(self, angle: float) -> float:
        raise NotImplementedError("Not implemented here")

    def discretize(self, refs):
        angles = [i * 2 * math.pi / refs for i in range(refs)]

        denominators = [
            (1 - self._a + self._a * math.cos(angle)) ** 2
            + self._a**2 * math.sin(angle) ** 2
            for angle in angles
        ]

        x_coords = [
            self._a * math.sin(angle) - self._a * math.sin(angle) / den
            for den, angle in zip(denominators, angles)
        ]
        y_coords = [
            self._a * math.cos(angle) + (1 - self._a + self._a * math.cos(angle)) / den
            for den, angle in zip(denominators, angles)
        ]
        mx, my = self.center
        return Polygon(
            [
                (mx + self._scale * x, my + self._scale * y)
                for x, y in zip(x_coords, y_coords)
            ]
        )



"""Transform for the pargeo package."""
import math
from typing import Any, Callable

from pargeo.domain import Domain, Transform
from pargeo.utils.geometry_utils import repeat
from pargeo.utils.typing_utils import Level, MultiPolygon, Polygon, SubDomain, Vector


class Periodic(Transform):
    """Periodic transform."""

    def __init__(
        self,
        levels: int | list[int] | str = "all",
        x_length: float | list[float] = float("inf"),
        y_length: float | list[float] = float("inf"),
        alpha: float | list[float] = 0,
    ):
        """Define the Periodic transform.

        Each level can have its own periodicity. If levels is "all" then all levels
        are treaded equally and the periodicity is applied to all levels.

        Args:
            levels: Levels to apply the periodicity. If "all" then the periodicity
                is applied to all levels. If a list of levels is given then the
                periodicity can be specified for each level individually.

            x_length: Periodic length in x direction. If "inf" then nothing
                happends in x direction. If a list is given then it must match the
                length of the levels list. Each entry is associated with the level
                with the corresponding index.

            y_length: Periodic length in y direction. Same as for
                `periodic_length_x`.

            alpha: Rotation angle in radians. Rotate the (x,y)-Plane ccw by alpha.
                NOTE: The polygons itself are not rotated, only the periodicity
                direction.
        """
        levels_ = levels if isinstance(levels, list) else []
        x_length = x_length if isinstance(x_length, list) else [x_length]
        y_length = y_length if isinstance(y_length, list) else [y_length]
        alpha = alpha if isinstance(alpha, list) else [alpha]
        self.all = False
        self.all_xlength = 0.0
        self.all_ylength = 0.0
        self.all_alpha = 0.0

        self.__lvl2xlength = {}
        self.__lvl2ylength = {}
        self.__lvl_2_alpha = {}

        # Validate input
        n_levels = 0
        if isinstance(levels, str):
            if levels == "all":
                if not (len(x_length) == 1 and len(y_length) == 1 and len(alpha) == 1):
                    raise ValueError(
                        "If levels is 'all' then x_length, y_length and alpha must be"
                        "floats."
                    )
                self.all = True
                n_levels = 1
                self.all_xlength = x_length[0]
                self.all_ylength = y_length[0]
                self.all_alpha = alpha[0]
            else:
                raise ValueError(
                    f"Levels specification as string must be 'all' but given {levels}."
                )
        elif isinstance(levels, int):
            levels_.append(levels)
            if not (len(x_length) == 1 and len(y_length) == 1 and len(alpha) == 1):
                raise ValueError(
                    "If levels is 'all' then x_length, y_length and alpha must be"
                    "floats."
                )
            self.all = False
            n_levels = 1
        elif isinstance(levels, list):
            n_levels = len(levels)

            x_length = x_length * n_levels if len(x_length) == 1 else x_length
            y_length = y_length * n_levels if len(y_length) == 1 else y_length
            alpha = alpha * n_levels if len(alpha) == 1 else alpha
            self.all = False
        else:
            raise ValueError(
                f"levels must be int, list[int] or 'all' but given {levels}."
            )
        if not self.all:
            self.__lvl2xlength = dict(zip(levels_, x_length))
            self.__lvl2ylength = dict(zip(levels_, y_length))
            self.__lvl_2_alpha = dict(zip(levels_, alpha))

    @property
    def affected_levels(self):
        """Return the levels affected by the periodicity transform."""
        return self.__lvl2xlength.keys()

    def __validate_input(
        self, level: int | str, x_length: float, y_length: float, alpha: float
    ):
        """Validate the input for the periodicity."""
        if not isinstance(level, int) and level != "all":
            raise ValueError(f"level must be int or `all` but given {level}.")
        if x_length <= 0:
            raise ValueError(f"x_length must be positive float but given {x_length}.")
        if y_length <= 0:
            raise ValueError(f"y_length must be positve float but given {y_length}.")
        if not isinstance(alpha, (float, int)):
            raise ValueError(f"alpha must be float but given {alpha}.")

    def update(self, level: int, x_length: float, y_length: float, alpha: float):
        """Update the periodic behavior of the transform.

        Note: If level is "all" then the periodicity is applied to all levels, meaning
        all levels are overwritten.
        """
        self.__validate_input(level, x_length, y_length, alpha)
        self.__lvl2xlength[level] = x_length
        self.__lvl2ylength[level] = y_length
        self.__lvl_2_alpha[level] = alpha

    def __call__(
        self, subdomain: SubDomain, level: int, domain: Domain, **kwargs: Any
    ) -> SubDomain:
        """Apply the periodicity to the polygon."""
        # Check if the polygon is affected by the periodicity.
        if level not in self.affected_levels and not self.all:
            return subdomain

        x_length = self.__lvl2xlength[level] if not self.all else self.all_xlength
        y_length = self.__lvl2ylength[level] if not self.all else self.all_ylength
        alpha = self.__lvl_2_alpha[level] if not self.all else self.all_alpha

        span: float = (
            domain.profile.bounds[2]
            - domain.profile.bounds[0]
            + domain.profile.bounds[3]
            - domain.profile.bounds[1]
        )
        n_reps_x = math.ceil(span / x_length)
        n_reps_y = math.ceil(span / y_length)

        return repeat(
            subdomain,
            ((math.cos(alpha), math.sin(alpha)), x_length, n_reps_x, True),
            ((math.sin(alpha), -math.cos(alpha)), y_length, n_reps_y, True),
            clipping_plane=domain.profile,
            grid_size=domain.grid_size,
        )


class Repeat(Transform):
    """Repeat transform."""

    def __init__(
        self,
        repeat_info: tuple[Vector, float, int, bool],
        repeat_info_2: tuple[Vector, float, int, bool] | None = None,
        excluded_levels: list[Level] = [],
    ) -> None:
        """Initialize a Repeat object.

        Args:
            repeat_info: Some info
            repeat_info_2: Some info
            excluded_levels: Levels to exclude from the transformation.
        """
        if repeat_info_2 is None:
            repeat_info_2 = ((0, 0), 0, 0, False)

        self.repeat_info = repeat_info
        self.repeat_info_2 = repeat_info_2
        self.excluded_levels = excluded_levels

    def __call__(self, subdomain: SubDomain, level: int, domain: Domain, **kwargs: Any):
        """Apply the repeat transform to the SubDomain.

        Args:
            subdomain: The SubDomain to transform.
            level: The Level of the transformation.
            domain: The Domain of the transformation.
            **kwargs: Additional keyword arguments for the transformation.

                clip (bool): If True, clips the repeated elements to the domain. Default is False.

        Returns:
            SubDomain: The transformed SubDomain.
        """
        if level in self.excluded_levels:
            return subdomain
        clip = kwargs.get("clip", True)
        clipping_plane = domain.profile if clip else None

        return repeat(
            subdomain,
            self.repeat_info,
            self.repeat_info_2,
            clipping_plane,
            domain.grid_size,
        )


class Diffeomorphism(Transform):
    """Diffeomorphism transform."""

    def __init__(
        self,
        mapping: Callable[[list[float], list[float]], tuple[list[float], list[float]]],
    ):
        """Generate a diffeomorphism transform. The mapping is applyed to the
        boundary of the polygon.

        Args:
            mapping: x_coords, y_coords -> x_mapped, y_mapped
        """
        self.__mapping = mapping

    def __call__(
        self, subdomain: SubDomain, level: int, domain: Domain, **kwargs: Any
    ) -> SubDomain:
        """Apply the diffeomorphism to the polygon."""
        if isinstance(subdomain, Polygon):
            return self.__map_polygon(subdomain)
        elif isinstance(subdomain, MultiPolygon):
            return MultiPolygon([self.__map_polygon(poly) for poly in subdomain])
        else:
            raise ValueError(f"Unknown polygon type {type(subdomain)}.")

    def __map_polygon(self, polygon: Polygon):
        """Transform the polygon boundary."""
        x, y = polygon.exterior.xy
        x_mapped, y_mapped = self.__mapping(x, y)

        shell = [(xx, yy) for xx, yy in zip(x_mapped, y_mapped)]

        holes = []
        if len(polygon.interiors) > 0:
            for hole in polygon.interiors:
                x, y = hole.exterior
                x_mapped, y_mapped = self.__mapping(x, y)
                hole_mapped = [(xx, yy) for xx, yy in zip(x_mapped, y_mapped)]
                holes.append(hole_mapped)
        return Polygon(shell, holes)



from typing import Mapping, Sequence, Union

import matplotlib.pyplot as plt

from pargeo.utils.constants import BACKGROUND_LEVEL
from pargeo.utils.typing_utils import Level, MultiPolygon, Polygon, SubDomain
from matplotlib.colors import Normalize

Color = Union[str, tuple[int, int, int], tuple[int, int, int, int]]


def plot_legend(
    holes: set[int],
    colormap: Mapping[Level, Color],
    background_level: Level = BACKGROUND_LEVEL,
):
    """Make a legend.

    Args:
        holes (set[int]): The holes.
        colormap (Mapping[Level, Color]): The color map.
        background_level (Level, optional): The background level. Defaults to BACKGROUND_LEVEL.
    """
    handles = []
    descriptions = []
    for lvl, color in colormap.items():
        handles.append(plt.Rectangle((0, 0), 1, 1, fc=color, edgecolor="black"))
        if lvl in holes:
            descriptions.append(f"{lvl}. level (hole)")
        else:
            if lvl == background_level:
                descriptions.append(f"Background (level = {background_level})")
            else:
                descriptions.append(f"{lvl}. level")
    plt.legend(
        handles,
        descriptions,
        loc="upper left",
        bbox_to_anchor=(0.95, 1),
    ).set_draggable(True)


def plot_polygon(
    polygon: Polygon,
    color: Color,
    boundary_color: Color,
    boundary_style: str,
    linewidth: float,
    show: bool = False,
):
    """Plot a polygon.

    Args:
        polygon (Polygon): The polygon to plot.
        color (Color): The fill color of the polygon.
        boundary_color (Color): The color of the boundary.
        boundary_style (str): The style of the boundary.
        linewidth (float): The width of the line.
        show (bool, optional): Whether to show the plot. Defaults to `False`.
    """
    x, y = polygon.exterior.xy
    plt.fill(x, y, color=color)

    # plot the exterior boundary
    if not boundary_color == "none":
        plt.plot(x, y, boundary_style, linewidth=linewidth, color=boundary_color)

    # plot interiors in white
    for interior in polygon.interiors:
        x, y = interior.xy
        plt.fill(x, y, color="white")
        plt.plot(x, y, "-", linewidth=1, color="black")
    if show:
        plt.show()


def plot_subdomain(
    subdomain: SubDomain,
    color="blue",
    boundary_color="black",
    boundary_style="-",
    linewidth=1.0,
    show: bool = True,
):
    """Plot the subdomain.

    Args:
        subdomain (SubDomain): The subdomain to plot.
        color (str, optional): The fill color of the subdomain. Defaults to `"blue"`.
        boundary_color (str, optional): The color of the boundary. Defaults to `"black`".
        boundary_style (str, optional): The style of the boundary. Defaults to `"-"`.
        linewidth (float, optional): The width of the line. Defaults to `1.0`.
        show (bool, optional): Whether to show the plot (using Matplotlib).
            Defaults to `True`.
    """
    if isinstance(subdomain, MultiPolygon):
        for polygon in subdomain.geoms:
            plot_polygon(polygon, color, boundary_color, boundary_style, linewidth)
    else:
        plot_polygon(subdomain, color, boundary_color, boundary_style, linewidth)
    if show:
        plt.show()


class DefaultColors:
    """Default colors for plotting."""

    boundary = "black"
    domain_boundary = "black"
    domain = "silver"
    hole = "white"
    interior_filling = "white"
    inter_hole_bound = "white"

    @staticmethod
    def get_color_map(
        levels: Sequence[int],
        holes: set[int],
        color_holes=False,
        background_level: int = BACKGROUND_LEVEL,
    ) -> Mapping[Level, Color]:
        """Get a color map for the levels.

        Args:
            levels (Sequence[int]): The levels.
            holes (set[int]): The holes.
            color_holes (bool, optional): Whether to color the holes. Defaults to `False`.
            background_level (int, optional): The background level. Defaults to `BACKGROUND_LEVEL`.
        """
        lvl2cl = dict()
        level_set = set(levels)
        if background_level in level_set:
            lvl2cl[background_level] = DefaultColors.domain
            level_set.remove(background_level)
        if len(level_set) > 0:
            norm = Normalize(vmin=min(level_set), vmax=max(level_set))
            for lvl in level_set:
                if lvl in holes and not color_holes:
                    lvl2cl[lvl] = DefaultColors.hole
                else:
                    lvl2cl[lvl] = plt.cm.cool(norm(lvl))  # type: ignore
        return lvl2cl
